/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package georouting.traversalAlgorithms;

import georouting.Graph;
import georouting.Node;
import georouting.PlanarizationAlgorithm;
import georouting.node.CFRNode;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.TreeMap;

/**
 *
 * @author Jordan
 */
public class CFRGF extends CFRG{
    private boolean mode_;  // true - DFlood | false - CFR
    
    public CFRGF(Graph g, Node s, Node f, PlanarizationAlgorithm pa){
        super(g, s, f, pa);
        mode_ = inArea(_current);
    }
    
    public CFRGF(Graph g, Node s, Node f, Node c, Node t, int id, ArrayList<Node> path, PlanarizationAlgorithm pa){
        super(g, s, f, c, t, id, path, pa);
        mode_ = inArea(_current);
    }
    
    @Override
    protected void advanceAlgorithm(){
        if(!done()){
//            System.out.println("Advance: " + this);
            if(!((CFRNode)target_).hasMate(this)){ 
                if(mode_){                      // Flood message - must be inArea -> inArea                               
                    moveTo(target_); 
                    visit(_current);
                
                    if(_current.neighborhood().size() > 1)
                        splitFlood(_previous);
                    else discard();
                }
                else if(inArea(target_)){       // CFR Message - CFR -> inArea
                    moveTo(target_);
                    visit(_current);
                    
                    // Send ACK back to CFR sender (opp dir)
                    CFRGF ack = new CFRGF(_container, _source, _destination, _current, _previous, id_, path_, _pAlg);
                    ack.mode_ = false;
                    ack._dir = _dir;
                    ((CFRNode)_current).receive(ack);
                    PostOffice.instance().add(ack);
                    
                    if(_current.neighborhood().size() > 1){
                        mode_ = true;   // Flood mode
                        splitFlood(_previous);
                    }                        
                    else discard();
                }
                else{                            // CFR Message - CFR -> CFR
                    moveTo(target_);
                    visit(_current);
                    
                    target_ = bestNeighbor();
                
                    if(juncture(_previous) || juncture(target_))
                        split(_previous); 
                }
            }
            else{
                PostOffice.instance().addPacket();
                broadcasts_++;
                discard();
            }
        }
    }
    
    @Override public void split(Node origin){
        if(mode_)
            splitFlood(origin);
        else
            splitCFR(origin);
    }
    
    public void splitFlood(Node origin){
        // Needs updating
//        System.out.println("Split - Flood");
        
        Iterator<Node> target_i = _current.neighborhood().iterator();
        
        Node target = target_ = null;
        CFRNode current = (CFRNode)_current;
        
        while(target_i.hasNext()){
            target = target_i.next();
            if(target != origin){
                if(inArea(target)){
                    target_ = target;
                    break;
                }
                else{
                    // Send CFR split
                    CFRGF fwdLeft  = new CFRGF(_container, _source, _destination, _current, target, ++seqNo_, path_, _pAlg);
                    CFRGF fwdRight = new CFRGF(_container, _source, _destination, _current, target, ++seqNo_, path_, _pAlg);
                    fwdRight.flipDirection();
                
                    fwdLeft.mode_ = false;
                    fwdRight.mode_ = false;
                    current.receive(fwdLeft);
                    current.receive(fwdRight);
                    PostOffice.instance().add(fwdLeft);
                    PostOffice.instance().add(fwdRight);
                }
            }
        }
        
        if(target_ != null){
            while(target_i.hasNext()){
                target = target_i.next();
                if(target != origin){
                    if(inArea(target)){
                        CFRGF spawn = new CFRGF(_container, _source, _destination, _current, target, id_, path_, _pAlg);
                        ((CFRNode)_current).receive(spawn);
                        PostOffice.instance().add(spawn);
                    }
                    else{
                        // Send CFR split
                        CFRGF fwdLeft  = new CFRGF(_container, _source, _destination, _current, target, ++seqNo_, path_, _pAlg);
                        CFRGF fwdRight = new CFRGF(_container, _source, _destination, _current, target, ++seqNo_, path_, _pAlg);
                        fwdRight.flipDirection();
                
                        fwdLeft.mode_ = false;
                        fwdRight.mode_ = false;
                        current.receive(fwdLeft);
                        current.receive(fwdRight);
                        PostOffice.instance().add(fwdLeft);
                        PostOffice.instance().add(fwdRight);
                    }
                }
            }
        }
        else
            discard();
    }
    
    public void splitCFR(Node origin){
//        System.out.println("Split - CFR");
        
        if(_current.neighborhood().size() > 1){
            TreeMap<Double, Node> sorted = new TreeMap<>();
            double zero = getAngle(origin);
                
            for(Node n : _current.neighborhood()){
                if(!sorted.containsValue(n)){
                    Double angle = getAngle(n) - zero;
                    if(angle < 0.0) angle += 2 * Math.PI;
                    sorted.put(angle, n);
                }
            }
            if(_dir == Direction.COUNTERCLOCKWISE)
                sorted.put(10.0, sorted.remove(0.0));
        
            Iterator<Node> first  = sorted.values().iterator();
            Iterator<Node> second = sorted.values().iterator();
            second.next();
        
            CFRNode current = (CFRNode)_current;

            while(second.hasNext()){
                Node left  = first.next();
                Node right = second.next();
                
                if(juncture(left) || juncture(right)){
                    CFRGF fwdLeft  = new CFRGF(_container, _source, _destination, _current, left, ++seqNo_, path_, _pAlg);
                    CFRGF fwdRight = new CFRGF(_container, _source, _destination, _current, right, ++seqNo_, path_, _pAlg);
                    fwdRight.flipDirection();
                
                    current.receive(fwdLeft);
                    current.receive(fwdRight);
                    PostOffice.instance().add(fwdLeft);
                    PostOffice.instance().add(fwdRight);
                }
            }
        }
    }
    
    protected void initiateFlood(){
        visit(_current);
        
        if(_current.neighborhood().size() > 0){           
            splitFlood(null);
            ((CFRNode)_current).receive(this);

            PostOffice.instance().merge();
        }
        else
            discard();
    }
    
    protected void initiateCFR(){
        Node left, right;
      
        if(_current.neighborhood().size() > 1){
            TreeMap<Double, Node> sorted = new TreeMap<>();
            double zero = getAngle(_destination);
      
            for(Node n : _current.neighborhood()){
                double angle = getAngle(n) - zero;
                if(angle < 0) angle += 2 * Math.PI;   
                sorted.put(angle, n);
            }
              
            if(sorted.firstKey() > 0.0){
                left = sorted.lastEntry().getValue();
                right = sorted.firstEntry().getValue();
            }
            else if(_current.neighborhood().size() > 2){
                // On S-D line and multiple secondary winners to choose from
                Node sd = sorted.firstEntry().getValue();
                sorted.remove(sorted.firstKey());

                if(2 * Math.PI - sorted.lastKey() < sorted.firstKey()){
                    left = sorted.lastEntry().getValue();
                    right = sd;
                }
                else{
                    left = sd;
                    right = sorted.firstEntry().getValue();
                }
            }
            else{
                // On S-D Line and single secondary
                left  = (sorted.lastKey() > Math.PI ? sorted.lastEntry().getValue() : sorted.firstEntry().getValue());
                right = (sorted.lastKey() > Math.PI ? sorted.firstEntry().getValue() : sorted.lastEntry().getValue());
            }
        }
        else{
            left = _current.neighborhood().get(0);
            right = left;
        }        
        
        target_ = left;
        _dir = Direction.CLOCKWISE;
        CFRGF mimic = new CFRGF(_container, _source, _destination, _current, right, id_, path_, _pAlg);
        mimic._dir = Direction.COUNTERCLOCKWISE;
    
        PostOffice.instance().generate(mimic);
        ((CFRNode)_current).receive(this);
        ((CFRNode)_current).receive(mimic);
        
        split(right);
        PostOffice.instance().merge();
    }
        
    @Override
    public boolean isMate(CFR message){
        if(mode_){
            return message.target_ == _current &&
               message.current() == target_;
        }
        else
            return super.isMate(message);
    }
    
    @Override
    public void initiate(){
        if(mode_)
            initiateFlood();
        else
            initiateCFR();
    }        
    
    public void flipMode(){
        mode_ = !mode_;
    }
    
    @Override
    public String toString(){
        return "{id: " + Integer.toString(id_) + ", c: " + Integer.toString(_current.id())
            + ", t: " + Integer.toString(target_.id()) + ", d: " + Integer.toString(_destination.id())
            + ", " + (_dir == Direction.CLOCKWISE ? "R" : "L") +", " + (mode_ ? "Flood}" : "CFR}");
    }
}
