/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package georouting.traversalAlgorithms;

import georouting.Graph;
import georouting.Node;
import georouting.PlanarizationAlgorithm;
import georouting.node.ImaginaryNode;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.KeyEvent;
import java.awt.geom.Ellipse2D;
import java.util.ArrayList;
import java.util.Iterator;

/**
 *
 * @author Jordan
 */
public class GOAFR extends FACE{
    private static final double R0_ = 1.4;          //
    private static final double R_  = Math.sqrt(2); //
    private static final double Om_ = 0.01;         // good > Om * bad goTo(best)
            
    private boolean     greedy_;
    private Ellipse2D   bound_;
    private Node        ui_;                        // where we switched
    private Node        faceMin_;                   // best
    private double      p_;                         // good counter
    private double      q_;                         // bad counter
    private boolean     crossed_;                   // ellipse mode        
    private ArrayList<Node> path_;
    
    public GOAFR(Graph g, Node s, Node d, PlanarizationAlgorithm pa){
        super(g, s, d, pa);
        greedy_ = true;
        double st = s.distance(d);
        bound_ = new Ellipse2D.Double(d.x() - st * R0_, d.y() - st * R0_,
                                     2 * st * R0_, 2 * st * R0_);
        path_ = new ArrayList<>();
        path_.add(s);
    }
    
    public ArrayList<Node> getPath(){
        // Extract loops from path
        Iterator<Node> it = path_.iterator();
        Node hop;
        int first, last;
        
        while(it.hasNext()){
            hop = it.next();
            first = path_.indexOf(hop);
            last = path_.lastIndexOf(hop);
            if(first != last){
                for(int i = first; i < last; i++){
                    it.remove();
                    it.next();
                }                    
            }            
        }
        path_.remove(0);
        return path_;
    }
    
    @Override
    public void advanceAlgorithm(){ 
        if(greedy_){
            Node min = _current;
            for(Node neighbor : _current.neighborhood())
                if(neighbor.distance(_destination) < min.distance(_destination))
                    min = neighbor;
                
            if(min != _current){
                moveTo(min);
                double RC = bound_.getWidth() / 2 / R_;
                if(_current.distance(_destination) < RC)
                    bound_.setFrame(_destination.x() - RC, _destination.y() - RC,
                                    RC * 2, RC * 2);
            }
            else
                enterFaceMode();            
        }
        // Face routing mode
        else {
            if(_current.distance(_destination) < faceMin_.distance(_destination)){
                faceMin_ = _current;
                p_++;
            }
            else q_++;
            
            if(p_ > Om_ * q_){
                advanceTo(faceMin_);
                greedy_ = true;                
            }
            else{
                Node target = bestNeighbor();
                if(crossesBound(target)){
                    if(!crossed_){                                             
                        crossed_ = true;
                        
                        _previous = target;
                        flipDirection();  
                        moveTo(bestNeighbor());
                    }
                    else{
                        double RC = bound_.getWidth() / 2 * R_;
                        bound_.setFrame(_destination.x() - RC, _destination.y() - RC,
                                        RC * 2, RC * 2);
                        
                        advanceTo(ui_);
                        enterFaceMode();
                    }
                }
                else
                    moveTo(target);
            }                
        }
    }
    
    private void enterFaceMode(){
        greedy_ = false;
        _previous = null;
        ui_ = _current;
        _X = new ImaginaryNode(_current);
        faceMin_ = _current;
        p_ = 0;
        q_ = 0;
        crossed_ = false;

        firstRun();
    }
    
    private void advanceTo(Node n){
        if(_current != n){
            moveTo(_previous);
            flipDirection();
            while(_current != n){
                moveTo(bestNeighbor());
//                System.out.println("\t" + Integer.toString(_current.id()));
            }
        }                    
    }
    
    private boolean crossesBound(Node n){
        return (bound_.contains(n.point()) != bound_.contains(_current.point()));
    }
    
    @Override
    public void moveTo(Node n){
        super.moveTo(n);
        path_.add(n);
    }
    
    @Override
    public void flipDirection(){
        if(_dir == Direction.CLOCKWISE)
            _dir = Direction.COUNTERCLOCKWISE;
        else
            _dir = Direction.CLOCKWISE; 
    }
    
    @Override
    public String toString(){
        String str = "S: " + Integer.toString(_source.id()) + ", D: " + Integer.toString(_destination.id())
                + ", C: " + Integer.toString(_current.id()) + ", Mode: ";
        if(greedy_)
            str += "Greedy";
        else {
            str += "Face, " + "p: " + Double.toString(p_) + ", q: " + Double.toString(q_);
            str += (crossed_ ? "-Cr" : "");
        }
        
        return str;
    }
    
    @Override
    public void visPaint(Graphics g)
    {
        super.visPaint(g);
        g.setColor(Color.cyan);
        g.drawOval((int)bound_.getX(), (int)bound_.getY(), (int)bound_.getWidth(), (int)bound_.getHeight());
        canvas().repaint();
    }
    
    @Override
    public void processKeyReleased(KeyEvent e)
    {
        super.processKeyReleased(e);
                
        if(done())
            System.out.println("Done - " + Integer.toString(_hops));
        else
            System.out.println(toString());
    }
    
    @Override
    public String name(){
        return "GOAFR+";
    }
}
