/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package georouting.traversalAlgorithms;

import georouting.TraversalAlgorithm;
import java.util.HashSet;
import java.util.Iterator;

/**
 * This class contains records of all messages currently in circulation, and
 * message statistics: total packets/messages sent, reception success, etc.
 */

public class PostOffice {
    static PostOffice singleton = new PostOffice();
    private HashSet<TraversalAlgorithm> archive;
    private HashSet<TraversalAlgorithm> inbox;
    private int messagesSent;
    private int packetsSent;
    private int received;
    
    static public PostOffice instance(){
        return singleton;
    }
    private PostOffice(){
        archive = new HashSet<>();
        inbox = new HashSet<>();
        messagesSent = 0;
        packetsSent = 0;
        received = 0;
    }
    
    public TraversalAlgorithm generate(TraversalAlgorithm msg){        
        archive.add(msg);
        messagesSent++;
        
        return msg;
    }
    
    public void add(TraversalAlgorithm msg){
        inbox.add(msg);
    }
    
    public void remove(TraversalAlgorithm message){
        archive.remove(message);
        inbox.remove(message);
    }
    
    public void advance(){
        Iterator<TraversalAlgorithm> msg_i = archive.iterator();
        
        while(msg_i.hasNext()){
            TraversalAlgorithm msg = msg_i.next();            
            msg.advance();
        }
        
        merge();
        
        msg_i = archive.iterator();
        
        while(msg_i.hasNext()){
            TraversalAlgorithm msg = msg_i.next();            
            if(msg.done()){
                msg_i.remove();
            }
        }
        
        msg_i = inbox.iterator();
        
        while(msg_i.hasNext()){
            TraversalAlgorithm msg = msg_i.next();            
            if(msg.done()){
                msg_i.remove();
            }
        }
    }
    
    public boolean deliver(TraversalAlgorithm msg){
        boolean success = msg.successful();
        if(success) received++;

        return success;
    }
    
    public void addPacket(){
        packetsSent++;
    }
    
    public int messages(){
        return messagesSent;
    }
    
    public int packets(){
        return packetsSent;
    }
    
    public float throughput(){
        return (float)received / messagesSent;
    }
    
    public float efficiency(){
       return (float)packetsSent / messagesSent;
    }
    
    public boolean isEmpty(){
        return archive.isEmpty();
    }
    
    public void reset(){
        archive.clear();
        inbox.clear();
        messagesSent = 0;
        packetsSent = 0;
        received = 0;
    }
    
    public void merge(){
        archive.addAll(inbox);
        inbox.clear();
    }
    
    @Override
    public String toString(){
        String stringRep = "/** PO **/\n";
        for(TraversalAlgorithm msg : archive){
            if(!msg.done())
                stringRep += msg.toString() + "\n";
        }
        stringRep += "/** End **/";
        
        return stringRep;
    }
}
